#######################################################################
# Copyright (c) 2025 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class InventoryDevice < ApplicationRecord
  belongs_to :sensor
  belongs_to :inventory_device_type_object, optional: true
  belongs_to :operating_system_object, optional: true
  belongs_to :hardware_object, optional: true
  belongs_to :criticality, optional: true

  has_many :mac_objects, dependent: :destroy

  has_and_belongs_to_many :incidents

  before_create :set_default_criticality

  accepts_nested_attributes_for :mac_objects, allow_destroy: true

  validate :ips_must_be_valid

  # ================================================================
  # Public cache methods
  # ================================================================
  def self.cache_key_for_user(user_id)
    "rb-webui:inventory-devices:#{user_id}"
  end
  
  # Save the current filters to Memcached
  def self.create_filters(user_id, search_params, name_param)
    filters = { search: search_params, name: name_param }
    Rails.cache.write(cache_key_for_user(user_id), filters, expires_in: 3.hours)
  end
  
  # Read the filters from Memcached
  def self.read_filters(user_id)
    Rails.cache.read(cache_key_for_user(user_id))
  end

  # Clear the filters from Memcached
   def self.clear_filters(user_id)
    Rails.cache.delete(cache_key_for_user(user_id))
  end

  def criticality_name
    self.criticality ? self.criticality.name : 'unset'
  end

  def to_asset(mac)
    type = self.inventory_device_type_object.name rescue nil
    os = self.operating_system_object.short_name rescue nil
    {
      id: self.id,
      name: self.name,
      mac_address: mac,
      type: type,
      operating_system: os
    }
  end
  
  def ips_must_be_valid
    return if ips.blank?

    ip_example = "(#{I18n.t('example_short')} #{I18n.t('ip_v4_example')} / #{I18n.t('ip_v6_example')} )"
  
    ips.each do |ip_value|
      unless ip_value.match?(Regexp.new(ip_pattern))
        errors.add(:ips, "#{ip_value} #{I18n.t('errors.messages.invalid')}\n#{ip_example}")
        next
      end
  
      begin
        IPAddr.new(ip_value)
      rescue IPAddr::InvalidAddressError, ArgumentError
        errors.add(:ips, "#{ip_value} #{I18n.t('errors.messages.invalid')}\n#{ip_example}")
      end
    end
  end

  def ip_pattern
    "^((([0-9]{1,3}\\.){3}[0-9]{1,3})|(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)))$"
  end

  private

  def set_default_criticality
    self.criticality ||= Criticality.find_or_create_by(name: "Unset")
  end
end