#######################################################################
# Copyright (c) 2025 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

require 'json'
require 'etc'

module DruidQueryLogger
  def debug(message = nil, &block)
    if message.present? && message.include?('DRUID')
      json_part = extract_json(message)
      save_to_file(json_part) if json_part
    end
    super
  end

  private

  MUTEX = Mutex.new

  def extract_json(message)
    json_match = message.match(/\{.*\}/m)
    json_match ? json_match[0] : nil
  end

  def save_to_file(json_part)
    file_path = Rails.root.join('log', 'druid_query_logging.log')
    check_and_delete_if_root(file_path)

    begin
      log_entry = JSON.parse(json_part)
      log_entry = log_entry.is_a?(Hash) ? log_entry.merge(
        'user' => safe_current_user_name,
        'creation_date' => Time.now
      ) : log_entry
    rescue JSON::ParserError
      return # Skip invalid JSON message
    end

    # Synchronize writing to the file using Mutex
    MUTEX.synchronize do
      File.open(file_path, 'a') do |f|
        f.puts(log_entry.to_json)
      end
    end
  end

  def safe_current_user_name
    User.current_user.name || 'Unknown user'
  rescue NoMethodError
    'Unknown user'
  end

  def check_and_delete_if_root(file_path)
    if File.exist?(file_path) && Rails.env.production?
      file_stat = File.stat(file_path)
      if file_stat.uid == Etc.getpwnam('root').uid && file_stat.gid == Etc.getgrnam('root').gid
        File.delete(file_path)
      end
    end
  end
end

Rails.logger.singleton_class.prepend(DruidQueryLogger)
