require 'rest-client'
require 'json'
require 'active_record'
require 'logger'
require 'yaml'
require 'pg'

class AppLogger
  def self.logger
    @logger ||= Logger.new(STDOUT).tap do |log|
      log.level = Logger::DEBUG
    end
  end
end

class DatabaseConfig
  def self.setup
    config = YAML.load_file("#{Rails.root}/config/database.yml")
    ActiveRecord::Base.establish_connection(config['production'])
  rescue StandardError => e
    AppLogger.logger.error("Database configuration error: #{e.message}")
    raise
  end

  def self.verify_connection
    ActiveRecord::Base.connection
    AppLogger.logger.info("Database connection established successfully.")
  rescue ActiveRecord::ConnectionNotEstablished => e
    AppLogger.logger.error("Failed to connect to the database: #{e.message}")
    raise
  end
end

class GLPIClient
  attr_reader :api_url

  def initialize(api_url, app_token)
    @api_url = api_url
    @app_token = app_token
    @logger = AppLogger.logger
  end

  def initiate_session(user_token)
    response = post_request("#{@api_url}/initSession", { user_token: user_token })
    JSON.parse(response.body)['session_token']
  rescue RestClient::ExceptionWithResponse => e
    @logger.error("Failed to start session: #{e.response}")
    raise
  end

  def end_session(session_token)
    post_request("#{@api_url}/killSession", {}, session_token)
  rescue RestClient::ExceptionWithResponse => e
    @logger.error("Failed to end session: #{e.response}")
  end

  def get_device_details(device_id, device_type, session_token)
    url = "#{@api_url}/#{device_type}/#{device_id}?with_networkports=true"
    response = get_request(url, session_token)
    device_details = JSON.parse(response.body)

    macs = []
    mac = get_network_port_mac(device_details, session_token)
    macs << mac unless mac.nil? || mac.empty?
    
    macs = macs + get_macs_from_device_network_card(device_details)

    operating_system = get_operating_system(device_id, session_token)
    ip_addresses = search_ip_addresses(device_id, session_token)

    { macs: macs, os: operating_system, ips: ip_addresses, details: device_details }
  rescue RestClient::ExceptionWithResponse => e
    @logger.error("Failed to retrieve details for device #{device_id}: #{e.response}")
    nil
  end

  private

  def headers(session_token = nil)
    {
      content_type: :json,
      accept: :json,
      'App-Token' => @app_token,
      'Session-Token' => session_token
    }
  end

  def post_request(url, payload, session_token = nil)
    RestClient.post(url, payload.to_json, headers(session_token))
  end

  def get_request(url, session_token)
    request_headers = headers(session_token)
    
    begin
      response = RestClient::Request.execute(
        method: :get,
        url: url,
        headers: request_headers
      )
      response
    rescue RestClient::ExceptionWithResponse => e
      @logger.error("Response Body: #{e.response.body}")
      raise
    end
  end

  def get_network_port_mac(device_details, session_token)
    network_ports_link = if device_details["links"]
      link = device_details["links"].find { |link| link["rel"] == "NetworkPort" }
      link ? link["href"] : nil
    end

    return nil unless network_ports_link

    begin
      uri = URI.parse(network_ports_link)
      uri.scheme = @api_url.split('://').first
      uri.host = URI.parse(@api_url).host
      network_ports_link = uri.to_s
    rescue URI::InvalidURIError => e
      @logger.error("Invalid URI: #{network_ports_link}")
      return nil
    end

    response = get_request(network_ports_link, session_token)
    network_ports_details = JSON.parse(response.body)

    port = network_ports_details.find { |p| p.is_a?(Hash) && p.key?('mac') }
    mac = port ? port['mac'] : nil
    mac
  rescue RestClient::ExceptionWithResponse => e
    @logger.error("Failed to retrieve network port details: #{e.response}")
    nil
  end

  def get_macs_from_device_network_card(device_details)
    macs = []
    if device_details['_devices'].is_a?(Hash) && device_details['_devices']['Item_DeviceNetworkCard'].is_a?(Hash)
      device_network_cards = device_details['_devices']['Item_DeviceNetworkCard']
      device_network_cards.each do |id, network_card|
        mac = network_card['mac']
        macs << mac unless mac.nil? || mac.strip.empty?
      end
    end

    macs
  end

  def get_operating_system(device_id, session_token)
    url = "#{@api_url}/search/Computer?criteria%5B0%5D%5Bfield%5D=2&criteria%5B0%5D%5Bsearchtype%5D=equals&criteria%5B0%5D%5Bvalue%5D=#{device_id}&forcedisplay%5B0%5D=45"
    response = get_request(url, session_token)
    search_results = JSON.parse(response.body)
  
    if search_results['data'].is_a?(Array) && !search_results['data'].empty?
      os = search_results['data'].first['45']
      if os
        return os
      end
    end

    nil
  rescue RestClient::ExceptionWithResponse => e
    log_error("Failed to retrieve operating system for device ID #{device_id}: #{e.response}")
  rescue StandardError => e
    logger.error("Unexpected error while retrieving OS for device ID #{device_id}: #{e.message}")
  end  

  def search_ip_addresses(device_id, session_token)
    search_url = "#{@api_url}/search/Computer"
    query_params = {
      "criteria[0][field]" => "2",  # Field ID for the device ID
      "criteria[0][searchtype]" => "equals",
      "criteria[0][value]" => device_id,
      "forcedisplay[0]" => "126",   # Field ID for IP address
      "itemtype" => "Computer",
      "start" => "0",
      "limit" => "50"
    }
    response = RestClient.get(search_url, headers(session_token).merge(params: query_params))
    results = JSON.parse(response.body)

    data = results['data']
    if data
      ip_addresses = data.flat_map { |device_data| device_data["126"] if device_data.is_a?(Hash) }.compact.uniq
    else
      ip_addresses = []
    end
    ip_addresses
  rescue RestClient::ExceptionWithResponse => e
    @logger.error("Failed to search IP addresses: #{e.response}")
    []
  end
end

class DeviceProcessor
  attr_reader :total_devices_retrieved, :total_devices_updated, :total_devices_inserted

  def initialize(client, domain_id)
    @client = client
    @domain_id = domain_id
    @logger = AppLogger.logger
    @total_devices_retrieved = 0
    @total_devices_updated = 0
    @total_devices_inserted = 0
  end

  def retrieve_device_details(session_token, device_type)
    start = 0
    batch_size = 50

    loop do
      range = "#{start}-#{start + batch_size - 1}"
      url = "#{@client.api_url}/#{device_type}?range=#{range}"
      response = @client.send(:get_request, url, session_token)
      devices = JSON.parse(response.body)
      @total_devices_retrieved += devices.size

      break if devices.empty?

      inventory_device_type_object = find_or_create_inventory_device_type_object(device_type)

      devices.each do |device|
        device_id = device["id"]
        @logger.info("Processing device ID: #{device_id}, Name: #{device['name']}")
        previously_existed = process_device(device, device_type, session_token, inventory_device_type_object)

        previously_existed ? @total_devices_updated += 1 : @total_devices_inserted += 1
      end


      content_range = response.headers[:content_range]
      if content_range
        total_devices = content_range.split('/').last.to_i
        start += batch_size
        break if start >= total_devices
      else
        @logger.warn("No Content-Range header found in response")
        break
      end

      sleep(10)
    end
  rescue RestClient::ExceptionWithResponse => e
    @logger.error("Failed to retrieve #{device_type} data: #{e.response}")
  end

  private

  def find_or_create_inventory_device_type_object(object_type)
    # For example: "NetworkEquipment" -> "Network Equipment"
    normalized_object_type = object_type.gsub(/([a-z])([A-Z])/, '\1 \2')

    InventoryDeviceTypeObject.find_or_create_by(name: normalized_object_type) do |type|
      type.description = "Type for #{normalized_object_type} made or updated by GLPI Integration"
    end
  end

  def find_or_initialize_mac_object(value, domain_id)
    mac_object = MacObject.find_by("value = ? AND sensor_id = ?", value, domain_id)
    mac_object ||= MacObject.new(value: value, sensor_id: domain_id)
  end

  def process_device(device, device_type, session_token, inventory_device_type_object)
    return false unless inventory_device_type_object

    device_details = @client.get_device_details(device["id"], device_type, session_token)
    unless device_details
      @logger.warn("Skipping device ID #{device['id']} due to missing details.")
      return false
    end

    mac_objects = []
    macs = device_details[:macs]
    os = device_details[:os]
    ips = device_details[:ips] || []
    comment = device["comment"]

    macs.each do |mac|
      mac_object = nil
      # Create MacObject
      ActiveRecord::Base.transaction do
        begin
          mac_object = find_or_initialize_mac_object(mac, @domain_id)

          mac_object_changed = false

          if mac_object.name != device["name"]
            mac_object.name = device["name"]
            mac_object_changed = true
          end

          if mac_object.comment != comment
            mac_object.comment = comment 
            mac_object_changed = true
          end

          if mac_object_changed
            mac_object.save!
          end
        rescue ActiveRecord::RecordInvalid => e
          @logger.error("Failed to save MacObject: #{e.record.errors.full_messages.join(", ")}")
        end
        mac_objects << mac_object 
      end
    end
    
    return false unless mac_objects.any?

    mac_objects_with_inventory_device = mac_objects.select { |mo| mo.inventory_device != nil }
    inventory_device = mac_objects_with_inventory_device.any? ? mac_objects_with_inventory_device.first.inventory_device : InventoryDevice.new

    ActiveRecord::Base.transaction do
      begin
      if inventory_device
        inventory_device.ips |= ips if ips.present?
        inventory_device.comments = comment if comment.present? && inventory_device.comments != comment
        inventory_device.mac_objects << mac_objects if mac_objects.present?
        inventory_device.sources |= ['GLPI']
        inventory_device.name = device["name"] if device["name"].present? && inventory_device.name != device["name"]
        inventory_device.operating_system_object = OsUtils.find_or_create_os(os)
        inventory_device.inventory_device_type_object = inventory_device_type_object

        inventory_device.updated_at = Time.current
        inventory_device.sensor_id = @domain_id
        return inventory_device.save!
      end
      rescue ActiveRecord::RecordInvalid => e
        @logger.error("Failed to save inventory_device: #{e.record.errors.full_messages.join(", ")}")
      end
    end
  rescue => e
    @logger.error("Error processing device ID #{device['id']}: #{e.message}")
    @logger.error(e.backtrace.join("\n"))
    false
  end
  false
end

module OsUtils
  def self.mapping
    @mapping ||= {}
  end

  def self.load_os_list
    json_file = Rails.root.join('files', 'objects', 'os_list.json')
    unless File.exist?(json_file)
      AppLogger.logger.error("OS list not found: #{json_file}")
      return nil
    end

    begin
      os_data = JSON.parse(File.read(json_file))
      mapping.clear
      os_data['operating_systems'].each do |os|
        product = os['product'].strip.downcase
        mapping[product] = {
          product: os['product'],
          family: os['family'],
          architecture: os['architecture'],
          vendor: os['vendor']
        }
      end
      mapping
    rescue JSON::ParserError => e
      AppLogger.logger.error("Failed to parse OS list JSON: #{e.message}")
      nil
    end
  end

  def self.find_or_create_os(device_os)
    return nil unless mapping

    parts = device_os ? device_os.strip.downcase.split : []
    product = parts.first || ''
    version = parts.second || ''

    base_attributes = mapping[product]
    if base_attributes
      os_attributes = base_attributes.merge(version: version)
    else
      os_attributes = {product: device_os, family: 'Other', architecture: '', vendor: '', version: version}
    end
    OperatingSystemObject.find_or_create_by(os_attributes)
  end
end

class GlpiInteraction
  def initialize(user_token, api_url, app_token, domain_id)
    @user_token = user_token
    @api_url = api_url
    @app_token = app_token
    @client = GLPIClient.new(api_url, app_token)
    @device_types = ["Computer", "NetworkEquipment", "Phone"]
    @domain_id = domain_id

    @logger = AppLogger.logger
  end

  def execute
    begin
      result = "Starting GLPI interaction.\n"
      @logger.info("Starting GLPI interaction.")
      DatabaseConfig.setup
      DatabaseConfig.verify_connection

      session_token = @client.initiate_session(@user_token)
      if session_token.nil?
        @logger.error("Failed to initiate session. Exiting...")
        raise "Session initiation failed"
      end

      unless OsUtils.load_os_list
        @logger.error('Failed to load OS list. Exiting...')
        raise 'OS list loading failed'
      end

      @device_types.each do |type|
        device_processor = DeviceProcessor.new(@client, @domain_id)
        device_processor.retrieve_device_details(session_token, type)
        result += "Batch complete for #{type}: Retrieved #{device_processor.total_devices_retrieved}, Updated #{device_processor.total_devices_updated}, Inserted #{device_processor.total_devices_inserted}\n"
      end

      result += "GLPI interaction completed successfully.\n"
    ensure
      @client.end_session(session_token) if session_token
      ActiveRecord::Base.connection.close
      @logger.info("Script is closing all connections.")
    end

    result
  rescue => e
    @logger.error("GLPI interaction failed: #{e.message}")
    @logger.error(e.backtrace.join("\n"))
    raise
  end
end
